program pcuboid;

{$APPTYPE CONSOLE}
{$B-}
{$N+}
{$E+}

uses
  SysUtils, Variants, Math, IniFiles, Classes, StrUtils;

type
  TPrime = record
    Prime,
    FactorA,
    FactorB: Int64;
    //K4P: 1..3;
  end;

  TPrimeArray = packed array of TPrime;

  TFactor = record
    Prime,
    FactorA,
    FactorB: Int64;
    Power: byte;
  end;

  TFactorArray = array of TFactor;

  TSquares = record
    FactorA,
    FactorB: Int64;
    byteA,
    byteB,
    sqrbA,
    sqrbB: Byte;
  end;

  TSquaresArray = array of TSquares;

  TResult = record
    A, B, C, D, E, F, G: Int64;
  end;

  TResultArray = array of TResult;

var
  ByteMatrix: packed array [1..255] of set of byte;
  P: TPrimeArray;
  S, Z, W: TFactorArray;
  Q, O: TSquaresArray;
  Res: TResultArray;
  A, B, C, D, E, F, G: Int64;
  ErrorCode: integer;
  filename: string;
  IniFile: TMemIniFile;
  TaskFinished: boolean;

const
  MinPrime = 5; //    4k+1
  IncPower = 2; //   
  MaxG: Extended = High(Int64)*1.0;
  Epcilon: Extended = 1e-3;

procedure SwitchInt64(var x, y: Int64);
var
  z: Int64;
begin
  if x<>y then begin
    z:=x;
    x:=y;
    y:=z;
  end;
end;

function CalcG(T: TFactorArray): int64;
  var i: integer;
  function IntPower(x: Int64; p: byte): Int64;
  var i: byte;
  begin
    Result:=1;
    for i:= 1 to p do
      Result:=Result*x;
  end;
begin
  Result:=1;
  i:=Low(T);
  while i<=High(T) do begin
    Result:=Result*IntPower(T[i].Prime, T[i].Power div 2);
    i:=i+1;
  end;
end;

procedure SwitchFactors(var T: TFactorArray; i, j: integer);
  var tmpf: TFactor;
begin
  if i<>j then begin
    tmpf:=T[i];
    T[i]:=T[j];
    T[j]:=tmpf;
  end;
end;

procedure SortFactorsByPrime(var T: TFactorArray; l, r: integer);
  var i, j, k: Integer;
begin
  repeat
    i:=l;
    j:=r;
    k:=(l+r) shr 1;
    repeat
      while T[i].Prime<T[k].Prime do inc(i);
      while T[j].Prime>T[k].Prime do dec(j);
      if i<=j then begin
        SwitchFactors(T, i, j);
        if k=i then k:=j
        else if k=j then k:=i;
        inc(i);
        dec(j);
      end;
    until i>j;
    if l<j then SortFactorsByPrime(T, l, j);
    l:=i;
  until i>=r;
end;

procedure SortFactorsByPower(var T: TFactorArray; l, r: integer);
  var i, j, k: Integer;
begin
  repeat
    i:=l;
    j:=r;
    k:=(l+r) shr 1;
    repeat
      while T[i].Power>T[k].Power do inc(i);
      while T[j].Power<T[k].Power do dec(j);
      if i<=j then begin
        SwitchFactors(T, i, j);
        if k=i then k:=j
        else if k=j then k:=i;
        inc(i);
        dec(j);
      end;
    until i>j;
    if l<j then SortFactorsByPower(T, l, j);
    l:=i;
  until i>=r;
end;

procedure SwitchSquares(var T: TSquaresArray; i, j: integer);
  var tmps: TSquares;
begin
  if i<>j then begin
    tmps:=T[i];
    T[i]:=T[j];
    T[j]:=tmps;
  end;
end;

procedure SortSquaresByFactorA(var T: TSquaresArray; l, r: integer);
  var i, j, k: Integer;
begin
  repeat
    i:=l;
    j:=r;
    k:=(l+r) shr 1;
    repeat
      while T[i].FactorA<T[k].FactorA do inc(i);
      while T[j].FactorA>T[k].FactorA do dec(j);
      if i<=j then begin
        SwitchSquares(T, i, j);
        if k=i then k:=j
        else if k=j then k:=i;
        inc(i);
        dec(j);
      end;
    until i>j;
    if l<j then SortSquaresByFactorA(T, l, j);
    l:=i;
  until i>=r;
end;

function LnFactors(var T: TFactorArray): Extended;
  var i: integer;
begin
  Result:=0;
  for i:=Low(T) to High(T) do
    Result:=Result+(T[i].Power div 2)*Ln(0.0+T[i].Prime);
end;

function FactorsValid(var T: TFactorArray): Boolean;
begin
  Result:=Ln(MaxG)>LnFactors(T);
end;

procedure DeleteFactor(var T: TFactorArray; i: Integer);
var j: integer;
begin
  for j:= i+1 to High(T) do
    T[j-1]:=T[j];
  SetLength(T, Length(T)-1);
end;

procedure FillSquaresFactorsByte(var T: TSquaresArray);
var i: Integer;
begin
  for i:= Low(T) to High(T) do begin
    T[i].byteA:=byte(T[i].FactorA);
    T[i].byteB:=byte(T[i].FactorB);
    T[i].sqrbA:=byte(T[i].byteA*T[i].byteA);
    T[i].sqrbB:=byte(T[i].byteB*T[i].byteB);
  end;
end;

procedure InitPrimes;
begin
  SetLength(P, 2);
  with P[0] do begin
    Prime:=2;
    FactorA:=1;
    FactorB:=1;
    //K4P:=2;
  end;
  with P[1] do begin
    Prime:=3;
    //K4P:=3;
  end;
end;

procedure InitByteMatrix;
var sq: set of byte;
  i, j, k: byte;
begin
  sq:=[];
  for i:= Low(byte) to High(byte) do
    sq:=sq+[byte(i*i)];
  for i:= Low(byte) to High(byte) do
    if i in sq then
      for j:= Low(byte) to High(byte) do
        if j in sq then begin
          k:=byte(i+j);
          if k in sq then ByteMatrix[k]:=ByteMatrix[k]+[i]+[j];
        end;
end;

function NextTaskParams(var T: TFactorArray; i: integer):Boolean;

  procedure AddFactor(var j: integer; aPower: byte);
    var k: integer;
  begin
    k:=Low(T);
    while (k<=High(T)) and (T[k].Prime<>P[j].Prime) do
      k:=k+1;
    if (k<=High(T)) and (T[k].Prime=P[j].Prime) then T[k].Power:=T[k].Power+aPower
    else begin
      SetLength(T, Length(T)+1);
      with T[High(T)] do begin
        Prime:=P[j].Prime;
        FactorA:=P[j].FactorA;
        FactorB:=P[j].FactorB;
        Power:=aPower;
      end;
    end;
  end;

  procedure EraseFactorsLessThan(var x: int64);
    var j: integer;
  begin
    SortFactorsByPrime(T, Low(T), High(T));
    j:=Low(T);
    while (j<=High(T)) and (T[j].Prime<x) do begin
      T[j].Power:=0;
      j:=j+1;
    end;
    SortFactorsByPower(T, Low(T), High(T));
    j:=High(T);
    while (j>=Low(T)) and (T[j].Power=0) do j:=j-1;
    SetLength(T, j+1);
  end;

begin
  Result:=True;
  if (i<=High(P)) then begin
    EraseFactorsLessThan(P[i].Prime);
    if (ln(0.0+P[i].Prime)<=Ln(MaxG)-LnFactors(T))
    then AddFactor(i, 2)
    else Result:=NextTaskParams(T, i+1);
  end
  else Result:=false;
end;

function SetTaskPrimes(var T: TFactorArray): Boolean;
  var
    i, j: Integer;
begin
  Result:=true;
  i:=Low(T);
  j:=Low(P);
  while Result and (i<=High(T)) do begin
    while (j<=High(P)) and (P[j].Prime<T[i].Prime) do
      j:=j+1;
    Result:=j<=High(P);
    if Result and (j<=High(P)) then
      if (T[i].Prime=P[j].Prime) (*and (P[j].K4P=1)*) then begin
        T[i].FactorA:=P[j].FactorA;
        T[i].FactorB:=P[j].FactorB;
        i:=i+1;
      end else begin
        Result:=NextTaskParams(T, j);
        i:=Low(T);
        j:=Low(P);
      end;
  end;
end;

procedure FindAllSquares(var V: TFactorArray);
  var
    i, j, k: integer;
    R: TSquaresArray;

  procedure DoubleCrossPairs(var T, U: TSquaresArray);
  var
    x, y: Int64;
    l: Integer;
  begin
    for l:= Low(T) to High(T) do begin
      x:=Abs(T[l].FactorA*V[i].FactorB-T[l].FactorB*V[i].FactorA);
      y:=T[l].FactorA*V[i].FactorA+T[l].FactorB*V[i].FactorB;
      if x > y then SwitchInt64(x, y);
      SetLength(U, Length(U)+1);
      with U[High(U)] do begin
        FactorA:=x;
        FactorB:=y;
      end;
      x:=Abs(T[l].FactorA*V[i].FactorA-T[l].FactorB*V[i].FactorB);
      y:=T[l].FactorA*V[i].FactorB+T[l].FactorB*V[i].FactorA;
      if x > y then SwitchInt64(x, y);
      SetLength(U, Length(U)+1);
      with U[High(U)] do begin
        FactorA:=x;
        FactorB:=y;
      end;
    end;
  end;

  procedure InsertNewPairs(var T, U: TSquaresArray);
  var
    m, n: integer;
    found, addnew: boolean;
  begin
    m:=Low(U);
    n:=Low(T);
    while m<=High(U) do begin
      if not((m>Low(U)) and (U[m].FactorA=U[m-1].FactorA)) then begin
        found:=false;
        addnew:=false;
        while not found and not addnew and (n<=High(T)) do begin
          found:=(T[n].FactorA=U[m].FactorA) and (T[n].FactorB=U[m].FactorB);
          if not found then
            if (T[n].FactorA<U[m].FactorA) then n:=n+1
            else addnew:=true;
        end;
        if not found then begin
          SetLength(T, Length(T)+1);
          with T[High(T)] do begin
            FactorA:=U[m].FactorA;
            FactorB:=U[m].FactorB;
          end;
        end;
      end;
      m:=m+1;
    end;
  end;

begin
  for i:= Low(V) to High(V) do begin
    for j:= 1 to V[i].Power do
      if (j mod 2)=1 then begin
        SetLength(Q, 0);
        SetLength(R, 0);
        DoubleCrossPairs(O, R);
        SortSquaresByFactorA(R, Low(R), High(R));
        InsertNewPairs(Q, R);
      end
      else begin
        for k:= Low(O) to High(O) do begin
          O[k].FactorA:=O[k].FactorA*V[i].Prime;
          O[k].FactorB:=O[k].FactorB*V[i].Prime;
        end;
        SetLength(R, 0);
        DoubleCrossPairs(Q, R);
        SortSquaresByFactorA(R, Low(R), High(R));
        InsertNewPairs(O, R);
        SortSquaresByFactorA(O, Low(O), High(O));
      end;
  end;
  Q:=Copy(O);
end;

procedure AddNewResult;
begin
  SetLength(Res, Length(Res)+1);
  Res[High(Res)].A:=A;
  Res[High(Res)].B:=B;
  Res[High(Res)].C:=C;
  Res[High(Res)].D:=D;
  Res[High(Res)].E:=E;
  Res[High(Res)].F:=F;
  Res[High(Res)].G:=G;
end;

function CheckSquares: Boolean;
  var i, j, maxj: Integer;
      intA, intB, intC, intD, intE, intF: Int64;
      extC, extD, maxB, minE: Extended;
      binA, binB: byte;
      found: Boolean;

  function CheckFactors(var x, y, z: int64): Boolean;
    var xw, yw, zw: longword;
  begin
    xw:=longword(abs(longint(x)));
    yw:=longword(abs(longint(y)));
    zw:=longword(abs(longint(z)));
    Result:=longword(xw*xw+yw*yw)=longword(zw*zw);
  end;

  function HexSqrRem(x: int64): byte;
  begin
    Result:=((byte(x) mod 8)*(byte(x) mod 8)) mod 8;
  end;

  function LegCorrespToHyp(x, y: byte): boolean;
  begin
    case x of
      0: Result:=(y in [0, 1, 4]);
      1: Result:=(y in [1]);
      4: Result:=(y in [0, 4])
      else Result:=null;
    end;
  end;

begin
  //Q     FactorA
  Result:=False;
  //ProgressPcnt:=0;
  maxj:=Length(Q);
  i:=Low(Q);
  G:=CalcG(S);
  while (i<=High(Q)-2) and (Q[i].FactorA*sqrt(2)<Q[i].FactorB) do begin
    //if maxj>1000 then ProgressPcnt:=byte(((i+1)*100 ) div maxj);
    intA:=Q[i].FactorA;
    binA:=Q[i].byteA mod 2;
    intF:=Q[i].FactorB;
    maxB:=intF/sqrt(2);
    minE:=IntA*sqrt(2);
    j:=i+1;
    while (intA>0) and (j<=High(Q)-1) and (Q[j].FactorA<maxB) and (minE<Q[j].FactorB) do begin
      if (Q[j].sqrbA in ByteMatrix[Q[i].sqrbB])
      and (Q[i].sqrbA in ByteMatrix[Q[j].sqrbB]) then begin
        binB:=Q[j].byteA mod 2;
        //         
        //        
        if (binA + binB <> 2) then begin
          intB:=Q[j].FactorA;
          intE:=Q[j].FactorB;
          //if CyclesCnt<High(CyclesCnt) then CyclesCnt:=CyclesCnt+1;
          extC:=sqrt(intF-intB*1.0)*sqrt(intF+intB*1.0);
          intC:=Round(extC);
          //         
          if (binA + binB + (byte(intC) mod 2) = 1) and SameValue(extC, intC, Epcilon) then begin
            extD:=sqrt(G-intC*1.0)*sqrt(G+intC*1.0);
            intD:=Round(extD);
            if SameValue(extD, intD, Epcilon) then begin
              A:=intA;
              B:=intB;
              C:=intC;
              D:=intD;
              E:=intE;
              F:=intF;
              found:=CheckFactors(A, B, D) and
                     CheckFactors(A, C, E) and
                     CheckFactors(B, C, F);
              if found then AddNewResult;
              Result:=Result or found;
            end;
          end;
        end;
      end;
      j:=j+1;
    end;
    maxj:=j;
    i:=i+1;
  end;
  //ProgressPcnt:=0;
end;

function CurrentTaskIsFinish: boolean;
var
  i, j: integer;
  DiffFound: boolean;
begin
  Result:=Length(Z)>0;
  if Result then begin
    //SortFactorsByPrime(S, Low(S), High(S));
    //S  Z      Prime
    i:=High(S);
    j:=High(Z);
    DiffFound:=false;
    Result:=false;
    while not Result and not DiffFound and (i>=Low(S)) and (j>=Low(Z)) do begin
      DiffFound:=(S[i].Prime<>Z[j].Prime);
      if DiffFound then
        Result:=S[i].Prime>Z[j].Prime
      else begin
        DiffFound:=(S[i].Power<>Z[j].Power);
        if DiffFound then
          Result:=S[i].Power>Z[j].Power;
        i:=i-1;
        j:=j-1;
      end;
    end;
    if not DiffFound then Result:=true;
  end;
end;

function PrimeFactorsAreCorrect(FactorsStr: string; var Prime, FactorA, FactorB: int64): boolean;
var
  DelimPos: Integer;
  FactorSqr: string;
begin
  DelimPos:=Pos('+', FactorsStr);
  Result:=DelimPos>0;
  if Result then begin
    FactorSqr:=LeftStr(FactorsStr, DelimPos-1);
    Result:=RightStr(FactorSqr, 2)='^2';
    if Result then begin
      Val(LeftStr(FactorSqr, Length(FactorSqr)-2), FactorA, ErrorCode);
      Result:=ErrorCode=0;
      if Result then begin
        FactorSqr:=Copy(FactorsStr, DelimPos+1, Length(FactorsStr));
        Result:=RightStr(FactorSqr, 2)='^2';
        if Result then begin
          Val(LeftStr(FactorSqr, Length(FactorSqr)-2), FactorB, ErrorCode);
          Result:=ErrorCode=0;
        end;
      end;
    end;
  end;
end;

function AddPrime(Prime, FactorA, FactorB: int64): boolean;
var i: Integer;
begin
  i:=0;
  while (i<=High(P)) and (P[i].Prime <> Prime) do
    i:=i+1;
  if i>High(P) then begin
    SetLength(P, Length(P)+1);
    P[High(P)].Prime:=Prime;
    //P[High(P)].K4P:=1;
    P[High(P)].FactorA:=FactorA;
    P[High(P)].FactorB:=FactorB;
    Result:=true;
  end
    else Result:=false;
end;

function AddFactor(var T: TFactorArray; Factor: int64; Power: byte): boolean;
var i: Integer;
begin
  i:=0;
  while (i<=High(P)) and (P[i].Prime <> Factor) do
    i:=i+1;
  if (i<=High(P)) then begin
    SetLength(T, Length(T)+1);
    T[High(T)].Prime:=Factor;
    //T[High(T)].K4P:=P[i].K4P;
    T[High(T)].FactorA:=P[i].FactorA;
    T[High(T)].FactorB:=P[i].FactorB;
    T[High(T)].Power:=Power;
    Result:=true;
  end
    else Result:=false;
end;

function SplitWorkUnitState(StateStr: string; var T: TFactorArray): boolean;
var
  DelimPos: Integer;
  Finished: Boolean;
  FactorStr: string;
  Factor: Int64;
  Power: Byte;
begin
  SetLength(T, 0);
  Finished:=False;
  Result:=True;
  repeat
    DelimPos:=Pos('*', StateStr);
    if DelimPos>0 then begin
      FactorStr:=LeftStr(StateStr, DelimPos-1);
      StateStr:=Copy(StateStr, DelimPos+1, Length(StateStr));
    end
    else begin
      FactorStr:=StateStr;
      StateStr:='';
    end;
    DelimPos:=Pos('^', FactorStr);
    if DelimPos>0 then begin
      Val(LeftStr(FactorStr, DelimPos-1), Factor, ErrorCode);
      if ErrorCode=0 then begin
        Val(Copy(FactorStr, DelimPos+1, Length(FactorStr)), Power, ErrorCode);
        if ErrorCode=0 then begin
          Result:=AddFactor(T, Factor, Power*2);
        end
        else Result:=False;
      end
      else Result:=False;
    end
      else Result:=False;
    if (Length(StateStr)=0) then
      if (Length(T)>0) then Finished:=true
      else Result:= false;
  until not Result or Finished;
  SortFactorsByPrime(T, Low(T), High(T));
end;

procedure InitFileTaskParams;
  var
    Sections, Keys: TStringList;
    Ini, Fin: string;
    Amt: integer;
    Prime, FactorA, FactorB: Int64;
    i: Byte;
  const
    WorkUnitSection = 'WU';
    PrimesSection = 'Primes';
begin
  try
    filename:=ParamStr(1);
    if ExtractFileDir(filename)='' then
      filename:=IncludeTrailingPathDelimiter(GetCurrentDir)+filename;
    IniFile:=TMemIniFile.Create(filename);
    Sections:=TStringList.Create;
    IniFile.ReadSections(Sections);
  except
    ErrorCode:=1;
  end;
  if ErrorCode=0 then try
    if Sections.IndexOf(WorkUnitSection)>=0 then begin
      Ini:=IniFile.ReadString(WorkUnitSection, 'Ini', '');
      Fin:=IniFile.ReadString(WorkUnitSection, 'Fin', '');
      Amt:=StrToInt(IniFile.ReadString(WorkUnitSection, 'Amt', ''));
    end
      else ErrorCode:=2;
  except
    ErrorCode:=2;
  end;
  if ErrorCode=0 then try
    if Sections.IndexOf(PrimesSection)>=0 then begin
      Keys:=TStringList.Create;
      IniFile.ReadSectionValues(PrimesSection, Keys);
      i:=0;
      while (ErrorCode=0) and (i<Keys.Count) do begin
        Val(Keys.Names[i], Prime, ErrorCode);
        if ErrorCode=0 then begin
          if PrimeFactorsAreCorrect(Keys.ValueFromIndex[i], Prime, FactorA, FactorB)
            and AddPrime(Prime, FactorA, FactorB) then ErrorCode:=0
          else ErrorCode:=3;
        end;
        i:=i+1;
      end;
    end
      else ErrorCode:=3;
  except
    ErrorCode:=3;
  end;
  if ErrorCode=0 then try
    if SplitWorkUnitState(Ini, S)
    and SplitWorkUnitState(Fin, Z)
    and (Ln(MaxG)>LnFactors(S))
    and (Ln(MaxG)>LnFactors(Z))
  then ErrorCode:=0
    else ErrorCode:=4;
  except
    ErrorCode:=4;
  end;
end;

procedure ShowResult;
var i: integer;
begin
  if Length(Res)=0 then writeln('Perfect cuboid is not found')
  else begin
    writeln('Perfect cuboid is found!');
    for i:=Low(Res) to High(Res) do begin
      writeln(i+1, '.');
      writeln('A = ', Res[i].A);
      writeln('B = ', Res[i].B);
      writeln('C = ', Res[i].C);
      writeln('D = ', Res[i].D);
      writeln('E = ', Res[i].E);
      writeln('F = ', Res[i].F);
      writeln('G = ', Res[i].G);
      writeln(Res[i].A, '^2 + ', Res[i].B, '^2 = ', Res[i].D, '^2');
      writeln(Res[i].A, '^2 + ', Res[i].C, '^2 = ', Res[i].E, '^2');
      writeln(Res[i].B, '^2 + ', Res[i].C, '^2 = ', Res[i].F, '^2');
      writeln(Res[i].A, '^2 + ', Res[i].B, '^2 + ', Res[i].C, '^2 = ', Res[i].G, '^2');
      writeln;
    end;
  end;
end;

begin
  InitByteMatrix;
  InitFileTaskParams;
  TaskFinished:=false;
  SetLength(Q, 0);
  while (ErrorCode=0) and not TaskFinished do begin
    //if MainForm.RefreshFreqEdit.Value=0 then Synchronize(ShowTaskParams);
    if (S[Low(S)].Prime=MinPrime) and (S[Low(S)].Power>=IncPower) and (Length(Q)>0) then begin
      //  ,    5- (    4k+1),
      //     Q
      SetLength(W, 1);
      W[Low(W)].Prime:=MinPrime;
      W[Low(W)].Power:=IncPower;
      SetTaskPrimes(W);
      O:=Copy(Q);
      FindAllSquares(W);
    end else begin
      //     G^2   
      SetLength(O, 1);
      O[Low(O)].FactorA:=0;
      O[Low(O)].FactorB:=1;
      FindAllSquares(S);
    end;
    FillSquaresFactorsByte(Q);
    //if MainForm.RefreshFreqEdit.Value=0 then Synchronize(ShowSquaresCnt);
    CheckSquares;
    //if MainForm.CheckBox.Checked then SaveSquares;
    //CheckedGCnt:=CheckedGCnt+1;
    //SquaresCnt:=SquaresCnt+Length(Q);
    //Duration:=Duration+Now-LastSave;
    //LastSave:=Now;
    //if MainForm.RefreshFreqEdit.Value=0 then Synchronize(ShowLastChecked);
    TaskFinished:=CurrentTaskIsFinish;
    if not TaskFinished then
      if not NextTaskParams(S, Low(P)) then ErrorCode:=5
      else SortFactorsByPrime(S, Low(S), High(S));
  end;
  if ErrorCode=0 then ShowResult;
end.
